@@ -5,98 +5,107 @@ Walks tables by primary key and searcher
 missing inserts/updates/deletes.
 """
 
-import sys, os, time, skytools
+import sys, os, skytools, subprocess
 
-from syncer import Syncer
+from londiste.syncer import Syncer
 
 __all__ = ['Repairer']
 
 def unescape(s):
+    """Remove copy escapes."""
     return skytools.unescape_copy(s)
 
-def get_pkey_list(curs, tbl):
-    """Get list of pkey fields in right order."""
-
-    oid = skytools.get_table_oid(curs, tbl)
-    q = """SELECT k.attname FROM pg_index i, pg_attribute k
-            WHERE i.indrelid = %s AND k.attrelid = i.indexrelid
-              AND i.indisprimary AND k.attnum > 0 AND NOT k.attisdropped
-            ORDER BY k.attnum"""
-    curs.execute(q, [oid])
-    list = []
-    for row in curs.fetchall():
-        list.append(row[0])
-    return list
-
-def get_column_list(curs, tbl):
-    """Get list of columns in right order."""
-
-    oid = skytools.get_table_oid(curs, tbl)
-    q = """SELECT a.attname FROM pg_attribute a
-            WHERE a.attrelid = %s
-              AND a.attnum > 0 AND NOT a.attisdropped
-            ORDER BY a.attnum"""
-    curs.execute(q, [oid])
-    list = []
-    for row in curs.fetchall():
-        list.append(row[0])
-    return list
-
 class Repairer(Syncer):
     """Walks tables in primary key order and checks if data matches."""
 
+    cnt_insert = 0
+    cnt_update = 0
+    cnt_delete = 0
+    total_src = 0
+    total_dst = 0
+    pkey_list = []
+    common_fields = []
+    apply_curs = None
 
-    def process_sync(self, tbl, src_db, dst_db):
+    def init_optparse(self, p=None):
+        """Initialize cmdline switches."""
+        p = super(Repairer, self).init_optparse(p)
+        p.add_option("--apply", action="store_true", help="apply fixes")
+        return p
+
+    def process_sync(self, src_tbl, dst_tbl, src_db, dst_db):
         """Actual comparision."""
 
+        apply_db = None
+
+        if self.options.apply:
+            apply_db = self.get_database('db', cache='applydb', autocommit=1)
+            self.apply_curs = apply_db.cursor()
+            self.apply_curs.execute("set session_replication_role = 'replica'")
+
         src_curs = src_db.cursor()
         dst_curs = dst_db.cursor()
 
-        self.log.info('Checking %s' % tbl)
+        self.log.info('Checking %s' % dst_tbl)
 
         self.common_fields = []
+        self.fq_common_fields = []
         self.pkey_list = []
-        copy_tbl = self.gen_copy_tbl(tbl, src_curs, dst_curs)
+        self.load_common_columns(src_tbl, dst_tbl, src_curs, dst_curs)
 
-        dump_src = tbl + ".src"
-        dump_dst = tbl + ".dst"
+        dump_src = dst_tbl + ".src"
+        dump_dst = dst_tbl + ".dst"
 
-        self.log.info("Dumping src table: %s" % tbl)
-        self.dump_table(tbl, copy_tbl, src_curs, dump_src)
+        self.log.info("Dumping src table: %s" % src_tbl)
+        self.dump_table(src_tbl, src_curs, dump_src)
         src_db.commit()
-        self.log.info("Dumping dst table: %s" % tbl)
-        self.dump_table(tbl, copy_tbl, dst_curs, dump_dst)
+        self.log.info("Dumping dst table: %s" % dst_tbl)
+        self.dump_table(dst_tbl, dst_curs, dump_dst)
         dst_db.commit()
         
-        self.log.info("Sorting src table: %s" % tbl)
+        self.log.info("Sorting src table: %s" % dump_src)
+        self.do_sort(dump_src, dump_src + '.sorted')
+        self.log.info("Sorting dst table: %s" % dump_dst)
+        self.do_sort(dump_dst, dump_dst + '.sorted')
 
-        s_in, s_out = os.popen4("sort --version")
-        s_ver = s_out.read()
-        del s_in, s_out
-        if s_ver.find("coreutils") > 0:
-            args = "-S 30%"
-        else:
-            args = ""
-        os.system("LC_ALL=C sort %s -T . -o %s.sorted %s" % (args, dump_src, dump_src))
-        self.log.info("Sorting dst table: %s" % tbl)
-        os.system("LC_ALL=C sort %s -T . -o %s.sorted %s" % (args, dump_dst, dump_dst))
-
-        self.dump_compare(tbl, dump_src + ".sorted", dump_dst + ".sorted")
+        self.dump_compare(dst_tbl, dump_src + ".sorted", dump_dst + ".sorted")
 
         os.unlink(dump_src)
         os.unlink(dump_dst)
         os.unlink(dump_src + ".sorted")
         os.unlink(dump_dst + ".sorted")
 
-    def gen_copy_tbl(self, tbl, src_curs, dst_curs):
-        self.pkey_list = get_pkey_list(src_curs, tbl)
-        dst_pkey = get_pkey_list(dst_curs, tbl)
+    def do_sort(self, src, dst):
+        p = subprocess.Popen(["sort", "--version"], stdout=subprocess.PIPE, stderr=subprocess.PIPE)
+        s_ver = p.communicate()[0]
+        del p
+
+        xenv = os.environ.copy()
+        xenv['LANG'] = 'C'
+        xenv['LC_ALL'] = 'C'
+
+        cmdline = ['sort', '-T', '.']
+        if s_ver.find("coreutils") > 0:
+            cmdline.append('-S')
+            cmdline.append('30%')
+        cmdline.append('-o')
+        cmdline.append(dst)
+        cmdline.append(src)
+        p = subprocess.Popen(cmdline, env = xenv)
+        if p.wait() != 0:
+            raise Exception('sort failed')
+
+    def load_common_columns(self, src_tbl, dst_tbl, src_curs, dst_curs):
+        """Get common fields, put pkeys in start."""
+
+        self.pkey_list = skytools.get_table_pkeys(src_curs, src_tbl)
+        dst_pkey = skytools.get_table_pkeys(dst_curs, dst_tbl)
         if dst_pkey != self.pkey_list:
             self.log.error('pkeys do not match')
             sys.exit(1)
 
-        src_cols = get_column_list(src_curs, tbl)
-        dst_cols = get_column_list(dst_curs, tbl)
+        src_cols = skytools.get_table_columns(src_curs, src_tbl)
+        dst_cols = skytools.get_table_columns(dst_curs, dst_tbl)
         field_list = []
         for f in self.pkey_list:
             field_list.append(f)
@@ -109,21 +118,24 @@ class Repairer(Syncer):
         self.common_fields = field_list
 
         fqlist = [skytools.quote_ident(col) for col in field_list]
+        self.fq_common_fields = fqlist
 
-        tbl_expr = "%s (%s)" % (skytools.quote_fqident(tbl), ",".join(fqlist))
+        cols = ",".join(fqlist)
+        self.log.debug("using columns: %s" % cols)
 
-        self.log.debug("using copy expr: %s" % tbl_expr)
+    def dump_table(self, tbl, curs, fn):
+        """Dump table to disk."""
+        cols = ','.join(self.fq_common_fields)
+        q = "copy %s (%s) to stdout" % (skytools.quote_fqident(tbl), cols)
 
-        return tbl_expr
-
-    def dump_table(self, tbl, copy_tbl, curs, fn):
         f = open(fn, "w", 64*1024)
-        curs.copy_to(f, copy_tbl)
+        curs.copy_expert(q, f)
         size = f.tell()
         f.close()
-        self.log.info('Got %d bytes' % size)
+        self.log.info('%s: Got %d bytes' % (tbl, size))
 
     def get_row(self, ln):
+        """Parse a row into dict."""
         if not ln:
             return None
         t = ln[:-1].split('\t')
@@ -133,6 +145,7 @@ class Repairer(Syncer):
         return row
 
     def dump_compare(self, tbl, src_fn, dst_fn):
+        """Dump + compare single table."""
         self.log.info("Comparing dumps: %s" % tbl)
         self.cnt_insert = 0
         self.cnt_update = 0
@@ -156,12 +169,12 @@ class Repairer(Syncer):
                 src_row = self.get_row(src_ln)
                 dst_row = self.get_row(dst_ln)
 
-                cmp = self.cmp_keys(src_row, dst_row)
-                if cmp > 0:
+                diff = self.cmp_keys(src_row, dst_row)
+                if diff > 0:
                     # src > dst
                     self.got_missed_delete(tbl, dst_row)
                     keep_src = 1
-                elif cmp < 0:
+                elif diff < 0:
                     # src < dst
                     self.got_missed_insert(tbl, src_row)
                     keep_dst = 1
@@ -182,6 +195,7 @@ class Repairer(Syncer):
                 self.cnt_insert, self.cnt_update, self.cnt_delete))
 
     def got_missed_insert(self, tbl, src_row):
+        """Create sql for missed insert."""
         self.cnt_insert += 1
         fld_list = self.common_fields
         fq_list = []
@@ -195,6 +209,7 @@ class Repairer(Syncer):
         self.show_fix(tbl, q, 'insert')
 
     def got_missed_update(self, tbl, src_row, dst_row):
+        """Create sql for missed update."""
         self.cnt_update += 1
         fld_list = self.common_fields
         set_list = []
@@ -215,6 +230,7 @@ class Repairer(Syncer):
         self.show_fix(tbl, q, 'update')
 
     def got_missed_delete(self, tbl, dst_row):
+        """Create sql for missed delete."""
         self.cnt_delete += 1
         whe_list = []
         for f in self.pkey_list:
@@ -223,16 +239,22 @@ class Repairer(Syncer):
         self.show_fix(tbl, q, 'delete')
 
     def show_fix(self, tbl, q, desc):
-        #self.log.warning("missed %s: %s" % (desc, q))
-        fn = "fix.%s.sql" % tbl
-        open(fn, "a").write("%s\n" % q)
+        """Print/write/apply repair sql."""
+        self.log.info("missed %s: %s" % (desc, q))
+        if self.apply_curs:
+            self.apply_curs.execute(q)
+        else:
+            fn = "fix.%s.sql" % tbl
+            open(fn, "a").write("%s\n" % q)
 
     def addeq(self, list, f, v):
+        """Add quoted SET."""
         vq = skytools.quote_literal(v)
         s = "%s = %s" % (f, vq)
         list.append(s)
 
     def addcmp(self, list, f, v):
+        """Add quoted comparision."""
         if v is None:
             s = "%s is null" % f
         else:
@@ -241,6 +263,7 @@ class Repairer(Syncer):
         list.append(s)
 
     def cmp_data(self, src_row, dst_row):
+        """Compare data field-by-field."""
         for k in self.common_fields:
             v1 = src_row[k]
             v2 = dst_row[k]
@@ -249,6 +272,7 @@ class Repairer(Syncer):
         return 0
 
     def cmp_value(self, v1, v2):
+        """Compare single field, tolerates tz vs notz dates."""
         if v1 == v2:
             return 0
 
