@@ -3,41 +3,83 @@
 
 import os
 from cStringIO import StringIO
-from skytools.quoting import quote_copy, quote_literal, quote_ident, quote_fqident
-import skytools.installer_config
+import skytools
+
+try:
+    import plpy
+except ImportError:
+    pass
 
 __all__ = [
     "fq_name_parts", "fq_name", "get_table_oid", "get_table_pkeys",
     "get_table_columns", "exists_schema", "exists_table", "exists_type",
+    "exists_sequence", "exists_temp_table", "exists_view",
     "exists_function", "exists_language", "Snapshot", "magic_insert",
     "CopyPipe", "full_copy", "DBObject", "DBSchema", "DBTable", "DBFunction",
     "DBLanguage", "db_install", "installer_find_file", "installer_apply_file",
+    "dbdict", "mk_insert_sql", "mk_update_sql", "mk_delete_sql",
 ]
 
+class dbdict(dict):
+    """Wrapper on actual dict that allows accessing dict keys as attributes."""
+    # obj.foo access
+    def __getattr__(self, k):
+        "Return attribute."
+        try:
+            return self[k]
+        except KeyError:
+            raise AttributeError(k)
+    def __setattr__(self, k, v):
+        "Set attribute."
+        self[k] = v
+    def __delattr__(self, k):
+        "Remove attribute."
+        del self[k]
+    def merge(self, other):
+        for key in other:
+            if key not in self:
+                self[key] = other[key]
 
 #
 # Fully qualified table name
 #
 
 def fq_name_parts(tbl):
-    "Return fully qualified name parts."
+    """Return fully qualified name parts.
+
+    >>> fq_name_parts('tbl')
+    ['public', 'tbl']
+    >>> fq_name_parts('foo.tbl')
+    ['foo', 'tbl']
+    >>> fq_name_parts('foo.tbl.baz')
+    ['foo', 'tbl.baz']
+    """
 
     tmp = tbl.split('.', 1)
     if len(tmp) == 1:
-        return ('public', tbl)
+        return ['public', tbl]
     elif len(tmp) == 2:
         return tmp
     else:
         raise Exception('Syntax error in table name:'+tbl)
 
 def fq_name(tbl):
-    "Return fully qualified name."
+    """Return fully qualified name.
+
+    >>> fq_name('tbl')
+    'public.tbl'
+    >>> fq_name('foo.tbl')
+    'foo.tbl'
+    >>> fq_name('foo.tbl.baz')
+    'foo.tbl.baz'
+    """
     return '.'.join(fq_name_parts(tbl))
 
 #
 # info about table
 #
 def get_table_oid(curs, table_name):
+    """Find Postgres OID for table."""
     schema, name = fq_name_parts(table_name)
     q = """select c.oid from pg_namespace n, pg_class c
            where c.relnamespace = n.oid
@@ -49,6 +91,7 @@ def get_table_oid(curs, table_name):
     return res[0][0]
 
 def get_table_pkeys(curs, tbl):
+    """Return list of pkey column names."""
     oid = get_table_oid(curs, tbl)
     q = "SELECT k.attname FROM pg_index i, pg_attribute k"\
         " WHERE i.indrelid = %s AND k.attrelid = i.indexrelid"\
@@ -58,6 +101,7 @@ def get_table_pkeys(curs, tbl):
     return map(lambda x: x[0], curs.fetchall())
 
 def get_table_columns(curs, tbl):
+    """Return list of column names for table."""
     oid = get_table_oid(curs, tbl)
     q = "SELECT k.attname FROM pg_attribute k"\
         " WHERE k.attrelid = %s"\
@@ -70,12 +114,14 @@ def get_table_columns(curs, tbl):
 # exist checks
 #
 def exists_schema(curs, schema):
+    """Does schema exists?"""
     q = "select count(1) from pg_namespace where nspname = %s"
     curs.execute(q, [schema])
     res = curs.fetchone()
     return res[0]
 
 def exists_table(curs, table_name):
+    """Does table exists?"""
     schema, name = fq_name_parts(table_name)
     q = """select count(1) from pg_namespace n, pg_class c
            where c.relnamespace = n.oid and c.relkind = 'r'
@@ -84,7 +130,28 @@ def exists_table(curs, table_name):
     res = curs.fetchone()
     return res[0]
 
+def exists_sequence(curs, seq_name):
+    """Does sequence exists?"""
+    schema, name = fq_name_parts(seq_name)
+    q = """select count(1) from pg_namespace n, pg_class c
+           where c.relnamespace = n.oid and c.relkind = 'S'
+             and n.nspname = %s and c.relname = %s"""
+    curs.execute(q, [schema, name])
+    res = curs.fetchone()
+    return res[0]
+
+def exists_view(curs, view_name):
+    """Does view exists?"""
+    schema, name = fq_name_parts(view_name)
+    q = """select count(1) from pg_namespace n, pg_class c
+           where c.relnamespace = n.oid and c.relkind = 'v'
+             and n.nspname = %s and c.relname = %s"""
+    curs.execute(q, [schema, name])
+    res = curs.fetchone()
+    return res[0]
+
 def exists_type(curs, type_name):
+    """Does type exists?"""
     schema, name = fq_name_parts(type_name)
     q = """select count(1) from pg_namespace n, pg_type t
            where t.typnamespace = n.oid
@@ -94,6 +161,7 @@ def exists_type(curs, type_name):
     return res[0]
 
 def exists_function(curs, function_name, nargs):
+    """Does function exists?"""
     # this does not check arg types, so may match several functions
     schema, name = fq_name_parts(function_name)
     q = """select count(1) from pg_namespace n, pg_proc p
@@ -110,18 +178,39 @@ def exists_function(curs, function_name, nargs):
     return res[0]
 
 def exists_language(curs, lang_name):
+    """Does PL exists?"""
     q = """select count(1) from pg_language
            where lanname = %s"""
     curs.execute(q, [lang_name])
     res = curs.fetchone()
     return res[0]
 
+def exists_temp_table(curs, tbl):
+    """Does temp table exists?"""
+    # correct way, works only on 8.2
+    q = "select 1 from pg_class where relname = %s and relnamespace = pg_my_temp_schema()"
+    curs.execute(q, [tbl])
+    tmp = curs.fetchall()
+    return len(tmp) > 0
+
 #
 # Support for PostgreSQL snapshot
 #
 
 class Snapshot(object):
-    "Represents a PostgreSQL snapshot."
+    """Represents a PostgreSQL snapshot.
+
+    Example:
+    >>> sn = Snapshot('11:20:11,12,15')
+    >>> sn.contains(9)
+    True
+    >>> sn.contains(11)
+    False
+    >>> sn.contains(17)
+    True
+    >>> sn.contains(20)
+    False
+    """
 
     def __init__(self, str):
         "Create snapshot from string."
@@ -158,38 +247,48 @@ def _gen_dict_copy(tbl, row, fields, qfields):
     tmp = []
     for f in fields:
         v = row.get(f)
-        tmp.append(quote_copy(v))
+        tmp.append(skytools.quote_copy(v))
     return "\t".join(tmp)
 
 def _gen_dict_insert(tbl, row, fields, qfields):
     tmp = []
     for f in fields:
         v = row.get(f)
-        tmp.append(quote_literal(v))
+        tmp.append(skytools.quote_literal(v))
     fmt = "insert into %s (%s) values (%s);"
     return fmt % (tbl, ",".join(qfields), ",".join(tmp))
 
 def _gen_list_copy(tbl, row, fields, qfields):
     tmp = []
     for i in range(len(fields)):
-        v = row[i]
-        tmp.append(quote_copy(v))
+        try:
+            v = row[i]
+        except IndexError:
+            v = None
+        tmp.append(skytools.quote_copy(v))
     return "\t".join(tmp)
 
 def _gen_list_insert(tbl, row, fields, qfields):
     tmp = []
     for i in range(len(fields)):
-        v = row[i]
-        tmp.append(quote_literal(v))
+        try:
+            v = row[i]
+        except IndexError:
+            v = None
+        tmp.append(skytools.quote_literal(v))
     fmt = "insert into %s (%s) values (%s);"
     return fmt % (tbl, ",".join(qfields), ",".join(tmp))
 
-def magic_insert(curs, tablename, data, fields = None, use_insert = 0):
-    """Copy/insert a list of dict/list data to database.
-    
+def magic_insert(curs, tablename, data, fields = None, use_insert = 0, quoted_table = False):
+    r"""Copy/insert a list of dict/list data to database.
+
     If curs == None, then the copy or insert statements are returned
     as string.  For list of dict the field list is optional, as its
     possible to guess them from dict keys.
+
+    Example:
+    >>> magic_insert(None, 'tbl', [[1, '1'], [2, '2']], ['col1', 'col2'])
+    'COPY public.tbl (col1,col2) FROM STDIN;\n1\t1\n2\t2\n\\.\n'
     """
     if len(data) == 0:
         return
@@ -210,15 +309,18 @@ def magic_insert(curs, tablename, data, fields = None, use_insert = 0):
         else:
             row_func = _gen_list_copy
 
-    qfields = [quote_ident(f) for f in fields]
-    qtablename = quote_fqident(tablename)
+    qfields = [skytools.quote_ident(f) for f in fields]
+    if quoted_table:
+        qtablename = tablename
+    else:
+        qtablename = skytools.quote_fqident(tablename)
 
     # init processing
     buf = StringIO()
     if curs == None and use_insert == 0:
         fmt = "COPY %s (%s) FROM STDIN;\n"
         buf.write(fmt % (qtablename, ",".join(qfields)))
- 
+
     # process data
     for row in data:
         buf.write(row_func(qtablename, row, fields, qfields))
@@ -245,18 +347,28 @@ def magic_insert(curs, tablename, data, fields = None, use_insert = 0):
 class CopyPipe(object):
     "Splits one big COPY to chunks."
 
-    def __init__(self, dstcurs, tablename = None, limit = 512*1024, cancel_func=None, sql_from = None):
+    def __init__(self, dstcurs, tablename = None, limit = 512*1024,
+                 sql_from = None):
         self.tablename = tablename
         self.sql_from = sql_from
         self.dstcurs = dstcurs
         self.buf = StringIO()
         self.limit = limit
-        self.cancel_func = None
+        #hook for new data, hook func should return new data
+        #def write_hook(obj, data):
+        #   return data
+        self.write_hook = None
+        #hook for flush, hook func result is discarded
+        # def flush_hook(obj):
+        #   return None
+        self.flush_hook = None
         self.total_rows = 0
         self.total_bytes = 0
 
     def write(self, data):
         "New data from psycopg"
+        if self.write_hook:
+            data = self.write_hook(self, data)
 
         self.total_bytes += len(data)
         self.total_rows += data.count("\n")
@@ -277,8 +389,8 @@ class CopyPipe(object):
     def flush(self):
         "Send data out."
 
-        if self.cancel_func:
-            self.cancel_func()
+        if self.flush_hook:
+            self.flush_hook(self)
 
         if self.buf.tell() <= 0:
             return
@@ -291,23 +403,55 @@ class CopyPipe(object):
         self.buf.seek(0)
         self.buf.truncate()
 
-def full_copy(tablename, src_curs, dst_curs, column_list = []):
+
+def full_copy(tablename, src_curs, dst_curs, column_list = [], condition = None,
+        dst_tablename = None, dst_column_list = None,
+        write_hook = None, flush_hook = None):
     """COPY table from one db to another."""
 
-    qtable = quote_fqident(tablename)
-    if column_list:
-        qfields = [quote_ident(f) for f in column_list]
-        hdr = "%s (%s)" % (qtable, ",".join(qfields))
+    # default dst table and dst columns to source ones
+    dst_tablename = dst_tablename or tablename
+    dst_column_list = dst_column_list or column_list[:]
+    if len(dst_column_list) != len(column_list):
+        raise Exception('src and dst column lists must match in length')
+
+    def build_qfields(cols):
+        if cols:
+            return ",".join([skytools.quote_ident(f) for f in cols])
+        else:
+            return "*"
+
+    def build_statement(table, cols):
+        qtable = skytools.quote_fqident(table)
+        if cols:
+            qfields = build_qfields(cols)
+            return "%s (%s)" % (qtable, qfields)
+        else:
+            return qtable
+
+    dst = build_statement(dst_tablename, dst_column_list)
+    if condition:
+        src = "(SELECT %s FROM %s WHERE %s)" % (build_qfields(column_list),
+                                                skytools.quote_fqident(tablename),
+                                                condition)
     else:
-        hdr = qtable
+        src = build_statement(tablename, column_list)
+
     if hasattr(src_curs, 'copy_expert'):
-        sql_to = "COPY %s TO stdout" % hdr
-        sql_from = "COPY %s FROM stdout" % hdr
+        sql_to = "COPY %s TO stdout" % src
+        sql_from = "COPY %s FROM stdin" % dst
         buf = CopyPipe(dst_curs, sql_from = sql_from)
+        buf.write_hook = write_hook
+        buf.flush_hook = flush_hook
         src_curs.copy_expert(sql_to, buf)
     else:
-        buf = CopyPipe(dst_curs, hdr)
-        src_curs.copy_to(buf, hdr)
+        if condition:
+            # regular psycopg copy_to generates invalid sql for subselect copy
+            raise Exception('copy_expert() is needed for conditional copy')
+        buf = CopyPipe(dst_curs, dst)
+        buf.write_hook = write_hook
+        buf.flush_hook = flush_hook
+        src_curs.copy_to(buf, src)
     buf.flush()
 
     return (buf.total_bytes, buf.total_rows)
@@ -323,11 +467,13 @@ class DBObject(object):
     sql = None
     sql_file = None
     def __init__(self, name, sql = None, sql_file = None):
+        """Generic dbobject init."""
         self.name = name
         self.sql = sql
         self.sql_file = sql_file
 
     def create(self, curs, log = None):
+        """Create a dbobject."""
         if log:
             log.info('Installing %s' % self.name)
         if self.sql:
@@ -339,47 +485,43 @@ class DBObject(object):
             sql = open(fn, "r").read()
         else:
             raise Exception('object not defined')
-        curs.execute(sql)
+        for stmt in skytools.parse_statements(sql):
+            #if log: log.debug(repr(stmt))
+            curs.execute(stmt)
 
     def find_file(self):
-        full_fn = None
-        if self.sql_file[0] == "/":
-            full_fn = self.sql_file
-        else:
-            dir_list = skytools.installer_config.sql_locations
-            for dir in dir_list:
-                fn = os.path.join(dir, self.sql_file)
-                if os.path.isfile(fn):
-                    full_fn = fn
-                    break
-
-        if not full_fn:
-            raise Exception('File not found: '+self.sql_file)
-        return full_fn
+        """Find install script file."""
+        return installer_find_file(self.sql_file)
 
 class DBSchema(DBObject):
     """Handles db schema."""
     def exists(self, curs):
+        """Does schema exists."""
         return exists_schema(curs, self.name)
 
 class DBTable(DBObject):
     """Handles db table."""
     def exists(self, curs):
+        """Does table exists."""
         return exists_table(curs, self.name)
 
 class DBFunction(DBObject):
     """Handles db function."""
     def __init__(self, name, nargs, sql = None, sql_file = None):
+        """Function object - number of args is significant."""
         DBObject.__init__(self, name, sql, sql_file)
         self.nargs = nargs
     def exists(self, curs):
+        """Does function exists."""
         return exists_function(curs, self.name, self.nargs)
 
 class DBLanguage(DBObject):
     """Handles db language."""
     def __init__(self, name):
+        """PL object - creation happens with CREATE LANGUAGE."""
         DBObject.__init__(self, name, sql = "create language %s" % name)
     def exists(self, curs):
+        """Does PL exists."""
         return exists_language(curs, self.name)
 
 def db_install(curs, list, log = None):
@@ -392,14 +534,16 @@ def db_install(curs, list, log = None):
                 log.info('%s is installed' % obj.name)
 
 def installer_find_file(filename):
+    """Find SQL script from pre-defined paths."""
     full_fn = None
     if filename[0] == "/":
         if os.path.isfile(filename):
             full_fn = filename
     else:
-        dir_list = ["."] + skytools.installer_config.sql_locations
-        for dir in dir_list:
-            fn = os.path.join(dir, filename)
+        import skytools.installer_config
+        dir_list = skytools.installer_config.sql_locations
+        for fdir in dir_list:
+            fn = os.path.join(fdir, filename)
             if os.path.isfile(fn):
                 full_fn = fn
                 break
@@ -409,12 +553,90 @@ def installer_find_file(filename):
     return full_fn
 
 def installer_apply_file(db, filename, log):
+    """Find SQL file and apply it to db, statement-by-statement."""
     fn = installer_find_file(filename)
     sql = open(fn, "r").read()
     if log:
         log.info("applying %s" % fn)
     curs = db.cursor()
     for stmt in skytools.parse_statements(sql):
-        log.debug(repr(stmt))
+        #log.debug(repr(stmt))
         curs.execute(stmt)
 
+#
+# Generate INSERT/UPDATE/DELETE statement
+#
+
+def mk_insert_sql(row, tbl, pkey_list = None, field_map = None):
+    """Generate INSERT statement from dict data.
+
+    >>> mk_insert_sql({'id': '1', 'data': None}, 'tbl')
+    "insert into public.tbl (data, id) values (null, '1');"
+    """
+
+    col_list = []
+    val_list = []
+    if field_map:
+        for src, dst in field_map.iteritems():
+            col_list.append(skytools.quote_ident(dst))
+            val_list.append(skytools.quote_literal(row[src]))
+    else:
+        for c, v in row.iteritems():
+            col_list.append(skytools.quote_ident(c))
+            val_list.append(skytools.quote_literal(v))
+    col_str = ", ".join(col_list)
+    val_str = ", ".join(val_list)
+    return "insert into %s (%s) values (%s);" % (
+                    skytools.quote_fqident(tbl), col_str, val_str)
+
+def mk_update_sql(row, tbl, pkey_list, field_map = None):
+    r"""Generate UPDATE statement from dict data.
+
+    >>> mk_update_sql({'id': 0, 'id2': '2', 'data': 'str\\'}, 'Table', ['id', 'id2'])
+    'update only public."Table" set data = E\'str\\\\\' where id = \'0\' and id2 = \'2\';'
+    """
+
+    if len(pkey_list) < 1:
+        raise Exception("update needs pkeys")
+    set_list = []
+    whe_list = []
+    pkmap = {}
+    for k in pkey_list:
+        pkmap[k] = 1
+        new_k = field_map and field_map[k] or k
+        col = skytools.quote_ident(new_k)
+        val = skytools.quote_literal(row[k])
+        whe_list.append("%s = %s" % (col, val))
+
+    if field_map:
+        for src, dst in field_map.iteritems():
+            if src not in pkmap:
+                col = skytools.quote_ident(dst)
+                val = skytools.quote_literal(row[src])
+                set_list.append("%s = %s" % (col, val))
+    else:
+        for col, val in row.iteritems():
+            if col not in pkmap:
+                col = skytools.quote_ident(col)
+                val = skytools.quote_literal(val)
+                set_list.append("%s = %s" % (col, val))
+    return "update only %s set %s where %s;" % (skytools.quote_fqident(tbl),
+            ", ".join(set_list), " and ".join(whe_list))
+
+def mk_delete_sql(row, tbl, pkey_list, field_map = None):
+    """Generate DELETE statement from dict data."""
+
+    if len(pkey_list) < 1:
+        raise Exception("delete needs pkeys")
+    whe_list = []
+    for k in pkey_list:
+        new_k = field_map and field_map[k] or k
+        col = skytools.quote_ident(new_k)
+        val = skytools.quote_literal(row[k])
+        whe_list.append("%s = %s" % (col, val))
+    whe_str = " and ".join(whe_list)
+    return "delete from only %s where %s;" % (skytools.quote_fqident(tbl), whe_str)
+
+if __name__ == '__main__':
+    import doctest
+    doctest.testmod()
